"""
BinPan Classes Main Module
"""

import pandas as pd
import numpy as np
from datetime import datetime
from decimal import Decimal as dd

from .market import get_prices_dic
from .quest import api_raw_get, api_raw_signed_get, check_weight
from .logs import Logs

base_url = 'https://api.binance.com'

stablecoins = ['PAX', 'TUSD', 'USDC', 'USDS', 'USDT', 'BUSD', 'DAI', 'UST', 'USDP', 'TRIBE', 'UST']

float_api_items = ['price', 'origQty', 'executedQty', 'cummulativeQuoteQty', 'stopLimitPrice', 'stopPrice', 'commission', 'qty',
                   'origQuoteOrderQty', 'makerCommission', 'takerCommission']
int_api_items = ['orderId', 'orderListId', 'transactTime', 'tradeId', 'transactionTime', 'updateTime', 'time']

exchange_logger = Logs(filename='./logs/exchange_logger.log', name='exchange_logger', info_level='INFO')


#########################
# exchange general data #
#########################


def get_exchange_info() -> dict:
    """
    Returns general exchange info from /api/v3/exchangeInfo endpoint.

    :return dict: dict_keys(['timezone', 'serverTime', 'rateLimits', 'exchangeFilters', 'symbols'])

    Example:

       ..code-block::

            {
              "timezone": "UTC",
              "serverTime": 1565246363776,
              "rateLimits": [
                {
                  //These are defined in the `ENUM definitions` section under `Rate Limiters (rateLimitType)`.
                  //All limits are optional
                }
              ],
              "exchangeFilters": [
                //These are the defined filters in the `Filters` section.
                //All filters are optional.
              ],
              "symbols": [
                {
                  "symbol": "ETHBTC",
                  "status": "TRADING",
                  "baseAsset": "ETH",
                  "baseAssetPrecision": 8,
                  "quoteAsset": "BTC",
                  "quotePrecision": 8,
                  "quoteAssetPrecision": 8,
                  "orderTypes": [
                    "LIMIT",
                    "LIMIT_MAKER",
                    "MARKET",
                    "STOP_LOSS",
                    "STOP_LOSS_LIMIT",
                    "TAKE_PROFIT",
                    "TAKE_PROFIT_LIMIT"
                  ],
                  "icebergAllowed": true,
                  "ocoAllowed": true,
                  "quoteOrderQtyMarketAllowed": true,
                  "allowTrailingStop": false,
                  "cancelReplaceAllowed": false,
                  "isSpotTradingAllowed": true,
                  "isMarginTradingAllowed": true,
                  "filters": [
                    //These are defined in the Filters section.
                    //All filters are optional
                  ],
                  "permissions": [
                     "SPOT",
                     "MARGIN"
                  ]
                }
              ]
            }
    """
    endpoint = '/api/v3/exchangeInfo'
    return api_raw_get(endpoint=endpoint,
                       weight=10)


def get_info_dic() -> dict:
    """
    Get the dictionary of each symbol with its information from the exchange.

    :return dict: Returns info for each symbol as keys in the dict.

    Example:

    .. code-block::

       from handlers import exchange

       info_dic = exchange.get_info_dic()

       info_dic['ETHUSDT']

        {'symbol': 'ETHUSDT',
         'status': 'TRADING',
         'baseAsset': 'ETH',
         'baseAssetPrecision': 8,
         'quoteAsset': 'USDT',
         'quotePrecision': 8,
         'quoteAssetPrecision': 8,
         'baseCommissionPrecision': 8,
         'quoteCommissionPrecision': 8,
         'orderTypes': ['LIMIT',
          'LIMIT_MAKER',
          'MARKET',
          'STOP_LOSS_LIMIT',
          'TAKE_PROFIT_LIMIT'],
         'icebergAllowed': True,
         'ocoAllowed': True,
         'quoteOrderQtyMarketAllowed': True,
         'allowTrailingStop': True,
         'cancelReplaceAllowed': True,
         'isSpotTradingAllowed': True,
         'isMarginTradingAllowed': True,
         'filters': [{'filterType': 'PRICE_FILTER',
           'minPrice': '0.01000000',
           'maxPrice': '1000000.00000000',
           'tickSize': '0.01000000'},
          {'filterType': 'PERCENT_PRICE',
           'multiplierUp': '5',
           'multiplierDown': '0.2',
           'avgPriceMins': 5},
          {'filterType': 'LOT_SIZE',
           'minQty': '0.00010000',
           'maxQty': '9000.00000000',
           'stepSize': '0.00010000'},
          {'filterType': 'MIN_NOTIONAL',
           'minNotional': '10.00000000',
           'applyToMarket': True,
           'avgPriceMins': 5},
          {'filterType': 'ICEBERG_PARTS', 'limit': 10},
          {'filterType': 'MARKET_LOT_SIZE',
           'minQty': '0.00000000',
           'maxQty': '6175.01628506',
           'stepSize': '0.00000000'},
          {'filterType': 'TRAILING_DELTA',
           'minTrailingAboveDelta': 10,
           'maxTrailingAboveDelta': 2000,
           'minTrailingBelowDelta': 10,
           'maxTrailingBelowDelta': 2000},
          {'filterType': 'MAX_NUM_ORDERS', 'maxNumOrders': 200},
          {'filterType': 'MAX_NUM_ALGO_ORDERS', 'maxNumAlgoOrders': 5}],
         'permissions': ['SPOT', 'MARGIN', 'TRD_GRP_004']}
    """
    return {k['symbol']: k for k in get_exchange_info()['symbols']}


###########
# account #
###########


def get_account_status(decimal_mode: bool,
                       api_key: str,
                       api_secret: str) -> dict:
    """
    Fetch account status detail.

    For Machine Learning limits, restrictions will be applied to account. If a user has been restricted by the ML system, they may
    check the reason and the duration by using the [/sapi/v1/account/status] endpoint

    :return dict: Example:

    .. code-block::

        { "data": "Normal" }

    """
    endpoint = '/sapi/v1/account/status'
    return api_raw_signed_get(endpoint=endpoint,
                              weight=1,
                              decimal_mode=decimal_mode,
                              api_key=api_key,
                              api_secret=api_secret)


def get_margin_bnb_interest_status(decimal_mode: bool,
                                   api_key: str,
                                   api_secret: str) -> dict:
    """
    Get BNB Burn Status (USER_DATA)

    GET /sapi/v1/bnbBurn (HMAC SHA256)

    Weight(IP): 1

    :return dict: Example

    .. code-block::

           {
               "spotBNBBurn":true,
               "interestBNBBurn": false
            }

    """
    endpoint = '/sapi/v1/bnbBurn'
    return api_raw_signed_get(endpoint=endpoint,
                              weight=1,
                              decimal_mode=decimal_mode,
                              api_key=api_key,
                              api_secret=api_secret)


#################
# weight limits #
#################


def get_exchange_limits(info_dict: dict = None) -> dict:
    """
    Binance manage several limits: RAW_REQUESTS, REQUEST_WEIGHT, and ORDERS rate limits.

    The headers for those limits, I assume that are:
    - RAW_REQUESTS: x-mbx-used-weight. Is cross all the api calls.
    - REQUEST_WEIGHT: Example: x-mbx-order-count-10s. Time interval limited requests.
    - ORDERS: Example: x-mbx-order-count-10s. Rate limit for orders.
    - X-SAPI-USED-IP-WEIGHT-1M: For sapi endpoint requests.

    Example response:

       .. code-block::

            {'x-mbx-used-weight-1m': 1200,
             'X-SAPI-USED-IP-WEIGHT-1M': 1200,
             'X-SAPI-USED-UID-WEIGHT-1M': 1200,
             'x-mbx-order-count-10s': 50,
             'x-mbx-order-count-1d': 160000,
             'x-mbx-used-weight': 6100}

    :return dict:
    """

    if not info_dict:
        info_dict = get_exchange_info()

    limits = info_dict['rateLimits']

    limits_dict = {}

    for i in limits:
        k3 = None
        if i['rateLimitType'].upper() == 'ORDERS':
            k1 = f"x-mbx-order-count-{i['intervalNum']}{i['interval'][0].lower()}"
            k2 = f"x-mbx-order-count-{i['intervalNum']}{i['interval'][0].lower()}"

        elif i['rateLimitType'].upper() == 'REQUEST_WEIGHT':
            k1 = f"x-mbx-used-weight-{i['intervalNum']}{i['interval'][0].lower()}"
            k2 = f"X-SAPI-USED-IP-WEIGHT-{i['intervalNum']}{i['interval'][0].upper()}"
            k3 = f"X-SAPI-USED-UID-WEIGHT-{i['intervalNum']}{i['interval'][0].upper()}"

        elif i['rateLimitType'].upper() == 'RAW_REQUESTS':
            k1 = "x-mbx-used-weight"
            k2 = "x-mbx-used-weight"
        else:
            raise Exception("BinPan Rate Limit not parsed")

        v = i['limit']

        limits_dict[k1] = v
        limits_dict[k2] = v
        if k3:
            limits_dict[k3] = v

    return limits_dict


##################
# symbol filters #
##################


def flatten_filter(filters: list) -> dict:
    """
    It flattens a dict to one level dictionary.

    :param list filters: A dict with API filter data.
    :return dict: A one level flattened dict with keys expanded with original sub-dicts.
    """
    ret = {}
    for f in filters:
        head = f['filterType']
        for k, v in f.items():
            if k != 'filterType':
                new_key = f"{head}_{k}"
                ret.update({new_key: v})
    return ret


def get_symbols_filters(info_dic: dict = None) -> dict:
    """
    Example:

    .. code-block:: Python

       exchange.get_filters('TLMBUSD', info_dic=info_dic)

       {'TLMBUSD': {'PRICE_FILTER_minPrice': '0.00001000',
          'PRICE_FILTER_maxPrice': '1000.00000000',
          'PRICE_FILTER_tickSize': '0.00001000',
          'PERCENT_PRICE_multiplierUp': '5',
          'PERCENT_PRICE_multiplierDown': '0.2',
          'PERCENT_PRICE_avgPriceMins': 5,
          'LOT_SIZE_minQty': '1.00000000',
          'LOT_SIZE_maxQty': '900000.00000000',
          'LOT_SIZE_stepSize': '1.00000000',
          'MIN_NOTIONAL_minNotional': '10.00000000',
          'MIN_NOTIONAL_applyToMarket': True,
          'MIN_NOTIONAL_avgPriceMins': 5,
          'ICEBERG_PARTS_limit': 10,
          'MARKET_LOT_SIZE_minQty': '0.00000000',
          'MARKET_LOT_SIZE_maxQty': '1181959.38584316',
          'MARKET_LOT_SIZE_stepSize': '0.00000000',
          'TRAILING_DELTA_minTrailingAboveDelta': 10,
          'TRAILING_DELTA_maxTrailingAboveDelta': 2000,
          'TRAILING_DELTA_minTrailingBelowDelta': 10,
          'TRAILING_DELTA_maxTrailingBelowDelta': 2000,
          'MAX_NUM_ORDERS_maxNumOrders': 200,
          'MAX_NUM_ALGO_ORDERS_maxNumAlgoOrders': 5}}

    :param dict info_dic: A BinPan exchange info dictionary

    :return dict: A dict with all flatten values.
    """
    if not info_dic:
        info_dic = get_info_dic()
    return {k: flatten_filter(v['filters']) for k, v in info_dic.items()}


def get_filters(symbol: str,
                info_dic: dict = None) -> dict:
    """
    For a symbol, get exchange filter conditions.

    :param str symbol: A Binance symbol.
    :param dict info_dic: BinPan exchange info dictionary. It's optional to avoid an API call. It's optional to avoid an API call.
    :return a dict: A dictionary with exclusively data for a symbol.

    Example:

    .. code-block::

        from handlers import exchange

        filters = exchange.get_filters('ETHBTC')

        filters

         {'ETHBTC': {'PRICE_FILTER_minPrice': '0.00000100',
          'PRICE_FILTER_maxPrice': '922327.00000000',
          'PRICE_FILTER_tickSize': '0.00000100',
          'PERCENT_PRICE_multiplierUp': '5',
          'PERCENT_PRICE_multiplierDown': '0.2',
          'PERCENT_PRICE_avgPriceMins': 5,
          'LOT_SIZE_minQty': '0.00010000',
          'LOT_SIZE_maxQty': '100000.00000000',
          'LOT_SIZE_stepSize': '0.00010000',
          'MIN_NOTIONAL_minNotional': '0.00010000',
          'MIN_NOTIONAL_applyToMarket': True,
          'MIN_NOTIONAL_avgPriceMins': 5,
          'ICEBERG_PARTS_limit': 10,
          'MARKET_LOT_SIZE_minQty': '0.00000000',
          'MARKET_LOT_SIZE_maxQty': '1135.26522780',
          'MARKET_LOT_SIZE_stepSize': '0.00000000',
          'TRAILING_DELTA_minTrailingAboveDelta': 10,
          'TRAILING_DELTA_maxTrailingAboveDelta': 2000,
          'TRAILING_DELTA_minTrailingBelowDelta': 10,
          'TRAILING_DELTA_maxTrailingBelowDelta': 2000,
          'MAX_NUM_ORDERS_maxNumOrders': 200,
          'MAX_NUM_ALGO_ORDERS_maxNumAlgoOrders': 5}}
    """
    if not info_dic:
        info_dic = get_info_dic()
    return {k: flatten_filter(v['filters']) for k, v in info_dic.items() if k == symbol}


####################################
# symbol characteristics filtering #
####################################


def filter_tradeable(info_dic: dict) -> dict:
    """
    Returns, from BinPan exchange info dictionary currently trading symbols.

    :param dict info_dic: BinPan exchange info dictionary. It's optional to avoid an API call.
    :return dict: BinPan exchange info dictionary, but, just with currently trading symbols.

    """
    return {k: v for k, v in info_dic.items() if v['status'].upper() == 'TRADING'}


def filter_spot(info_dic: dict) -> dict:
    """
    Returns, from BinPan exchange info dictionary currently SPOT symbols.

    :param dict info_dic: BinPan exchange info dictionary. It's optional to avoid an API call.
    :return dict: BinPan exchange info dictionary, but, just with currently SPOT symbols.

    """
    return {k: v for k, v in info_dic.items() if 'SPOT' in v['permissions'] or 'spot' in v['permissions']}


def filter_margin(info_dic: dict) -> dict:
    """
    Returns, from BinPan exchange info dictionary currently MARGIN symbols.

    :param dict info_dic: BinPan exchange info dictionary. It's optional to avoid an API call.
    :return dict: BinPan exchange info dictionary, but, just with currently MARGIN symbols.

    """
    return {k: v for k, v in info_dic.items() if 'MARGIN' in v['permissions'] or 'margin' in v['permissions']}


def filter_not_margin(symbols: list = None,
                      info_dic: dict = None) -> list:
    """
    Returns, from BinPan exchange info dictionary currently NOT MARGIN symbols.

    :param list symbols: A list of symbols to apply filter. Optional.
    :param dict info_dic: BinPan exchange info dictionary. It's optional to avoid an API call.
    :return list: BinPan exchange info dictionary, but, just with currently NOT MARGIN symbols.

    """
    if not info_dic:
        info_dic = get_info_dic()

    permissions_dic = {k: v['permissions'] for k, v in info_dic.items()}

    if symbols:
        return [s for s, p in permissions_dic.items() if 'MARGIN' in p and s in symbols]
    else:
        return [s for s, p in permissions_dic.items() if 'MARGIN' in p]


def filter_leveraged_tokens(info_dic: dict) -> dict:
    """
    Returns, from BinPan exchange info dictionary currently NOT LEVERAGED symbols.

    :param dict info_dic: BinPan exchange info dictionary. It's optional to avoid an API call.
    :return dict: BinPan exchange info dictionary, but, just with currently NOT LEVERAGED symbols.

    """
    return {k: v for k, v in info_dic.items() if all(lev not in k for lev in ['UP', 'DOWN', 'BULL', 'BEAR'])}


def filter_legal(legal_coins: list,
                 info_dic: dict) -> dict:
    """
    Returns, from BinPan exchange info dictionary currently trading symbols not using legal Fiat money.

    :param list legal_coins: List of legal coins, Fiat coins. 
    :param dict info_dic: BinPan exchange info dictionary. It's optional to avoid an API call.
    :return dict: BinPan exchange info dictionary, but, without fiat legal symbols.

    """
    bases = get_bases_dic(info_dic=info_dic)
    quotes = get_quotes_dic(info_dic=info_dic)
    return {k: v for k, v in info_dic.items() if (not bases[k] in legal_coins) and (not quotes[k] in legal_coins)}


#################
# Exchange Data #
#################


def get_precision(info_dic: dict = None) -> dict:
    """
    Gets a dictionary with decimal positions each symbol.
    :param dict info_dic: BinPan exchange info dictionary. It's optional to avoid an API call.
    :return dict: dictionary with decimal positions each symbol.
    """
    if not info_dic:
        info_dic = get_info_dic()
    return {k: {'baseAssetPrecision': v['baseAssetPrecision'],
                'quoteAssetPrecision': v['quoteAssetPrecision'],
                'baseCommissionPrecision': v['baseCommissionPrecision'],
                'quoteCommissionPrecision': v['quoteCommissionPrecision']}
            for k, v in info_dic.items()}


def get_orderTypes_and_permissions(info_dic: dict = None) -> dict:
    """
    Gets a dictionary with a list of order types suppoerted each symbol.
    :param dict info_dic: BinPan exchange info dictionary. It's optional to avoid an API call.
    :return dict: dictionary with a list of order types suppoerted each symbol.
    """
    if not info_dic:
        info_dic = get_info_dic()
    return {k: {'orderTypes': v['orderTypes'], 'permissions': v['permissions']} for k, v in info_dic.items()}


def get_fees_dict(decimal_mode: bool,
                  api_key: str,
                  api_secret: str,
                  symbol: str = None
                  ) -> dict:
    """
    Returns fees for a symbol or for every symbol if not passed a symbol.

    :param str symbol: Optional to request just one symbol instead of all.
    :param str api_key: A BinPan encrypted in BinPan's way api key. Do not use unencrypted api values, import it from  secret.py file.
    :param str api_secret: A BinPan encrypted in BinPan's way api secret. Do not use unencrypted api values, import it from  secret.py file.
    :param bool decimal_mode: Fixes Decimal return type.
    :return dict: A dict with maker and taker fees.
    """
    endpoint = '/sapi/v1/asset/tradeFee'
    if symbol:
        symbol = symbol.upper()
    ret = api_raw_signed_get(endpoint,
                             params={'symbol': symbol},
                             weight=1,
                             decimal_mode=decimal_mode,
                             api_key=api_key,
                             api_secret=api_secret)
    if decimal_mode:
        return {i['symbol']: {'makerCommission': dd(i['makerCommission']),
                              'takerCommission': dd(i['takerCommission'])} for i in ret}
    else:
        return {i['symbol']: {'makerCommission': float(i['makerCommission']),
                              'takerCommission': float(i['takerCommission'])} for i in ret}


def get_fees(decimal_mode: bool,
             api_key: str,
             api_secret: str,
             symbol: str = None) -> pd.DataFrame:
    """
    Returns fees for a symbol or for every symbol if not passed.

    :param bool decimal_mode: Fixes Decimal return type and operative.
    :param str api_key: A BinPan encrypted in BinPan's way api key. Do not use unencrypted api values, import it from  secret.py file.
    :param str api_secret: A BinPan encrypted in BinPan's way api secret. Do not use unencrypted api values, import it from  secret.py file.
    :param str symbol: Optional to request just one symbol instead of all.
    :return pd.DataFrame: A pandas dataframe with all the fees applied each symbol.
    """
    ret = get_fees_dict(symbol=symbol, decimal_mode=decimal_mode, api_key=api_key, api_secret=api_secret)
    return pd.DataFrame(ret).transpose()


def get_system_status():
    """
    Fetch system status.

    Weight(IP): 1

    :return dict: As shown in example.

    Example:

    .. code-block::

        {
           "status": 0,              // 0: normal，1：system maintenance
            "msg": "normal"           // "normal", "system_maintenance"
        }

    """
    return api_raw_get(endpoint='/sapi/v1/system/status',
                       weight=1)['msg']


def get_coins_and_networks_info(decimal_mode: bool,
                                api_key: str,
                                api_secret: str) -> tuple:
    """
    Get information of coins (available for deposit and withdraw) for user.

    GET /sapi/v1/capital/config/getall (HMAC SHA256)

    Weight(IP): 10

    :return pd.DataFrame, pd.DataFrame:
    """

    ret = api_raw_signed_get(endpoint='/sapi/v1/capital/config/getall',
                             weight=10, decimal_mode=decimal_mode,
                             api_key=api_key,
                             api_secret=api_secret)
    networks = []
    coins = []

    for i in ret:
        network_list = i.pop('networkList', None)
        if network_list:
            networks += network_list
        coins.append(i)
    # networks = list(set(networks))

    coins_df = pd.DataFrame(coins).drop_duplicates().set_index('coin')
    for col in coins_df.columns:
        coins_df[col] = pd.to_numeric(arg=coins_df[col], downcast='integer', errors='ignore')

    networks_df = pd.DataFrame(networks).drop_duplicates().set_index('coin')
    for col in networks_df.columns:
        networks_df[col] = pd.to_numeric(arg=networks_df[col], downcast='integer', errors='ignore')

    return coins_df.sort_index(), networks_df.sort_index()


def get_coins_info_list(decimal_mode: bool,
                        api_key: str,
                        api_secret: str,
                        coin: str = None) -> list:
    """
    Bring all coins exchange info in a list if no one is specified.

    Returns a list of dictionaries, one for each currency.

    :param bool decimal_mode: Fixes Decimal return type and operative.
    :param str api_key: A BinPan encrypted in BinPan's way api key. Do not use unencrypted api values, import it from  secret.py file.
    :param str api_secret: A BinPan encrypted in BinPan's way api secret. Do not use unencrypted api values, import it from  secret.py file.
    :param str coin: Limit response to a coin.
    :return list: A list of dictionaries each coin.
    """

    endpoint = '/sapi/v1/capital/config/getall'
    check_weight(weight=10, endpoint=endpoint)
    ret = api_raw_signed_get(endpoint=endpoint,
                             decimal_mode=decimal_mode,
                             api_key=api_key,
                             api_secret=api_secret)
    if not coin:
        return ret
    else:
        return [c for c in ret if c['coin'].upper() == coin.upper()]


def get_coins_info_dic(decimal_mode: bool,
                       api_key: str,
                       api_secret: str,
                       coin: str = None) -> dict:
    """
    Useful managing coins info in a big dictionary with coins as keys.

    :param bool decimal_mode: Fixes Decimal return type and operative.
    :param str api_key: A BinPan encrypted in BinPan's way api key. Do not use unencrypted api values, import it from  secret.py file.
    :param str api_secret: A BinPan encrypted in BinPan's way api secret. Do not use unencrypted api values, import it from  secret.py file.
    :param str coin: Limit response to a coin.
    :return list: A dictionary with each coin data as value.
    """
    coins_data_list = get_coins_info_list(coin=coin,
                                          decimal_mode=decimal_mode,
                                          api_key=api_key,
                                          api_secret=api_secret)
    return {c['coin']: c for c in coins_data_list}


def get_legal_coins(decimal_mode: bool,
                    api_key: str,
                    api_secret: str,
                    coins_dic: dict = None) -> list:
    """
    Fetch coins containing isLegalMoney=true

    :param bool decimal_mode: Fixes Decimal return type and operative.
    :param str api_key: A BinPan encrypted in BinPan's way api key. Do not use unencrypted api values, import it from  secret.py file.
    :param str api_secret: A BinPan encrypted in BinPan's way api secret. Do not use unencrypted api values, import it from  secret.py file.
    :param dict coins_dic: Avoid fetching the API by passing a dict with coins data.
    :return list: A list with coins names.
    """
    if not coins_dic:
        coins_dic = get_coins_info_dic(decimal_mode=decimal_mode,
                                       api_key=api_key,
                                       api_secret=api_secret)
    return [coin for coin, data in coins_dic.items() if data['isLegalMoney'] is True]


def get_leveraged_coins(decimal_mode: bool,
                        api_key: str,
                        api_secret: str,
                        coins_dic: dict = None) -> list:
    """
    Search for Binance leveraged coins by searching UP or DOWN before an existing coin, examples:

        .. code-block:: python

            ['1INCHDOWN', '1INCHUP', 'AAVEDOWN', 'AAVEUP', 'ADADOWN', ... ]

    :param bool decimal_mode: Fixes Decimal return type and operative.
    :param str api_key: A BinPan encrypted in BinPan's way api key. Do not use unencrypted api values, import it from  secret.py file.
    :param str api_secret: A BinPan encrypted in BinPan's way api secret. Do not use unencrypted api values, import it from  secret.py file.
    :param dict coins_dic: Avoid fetching the API by passing a dict with coins data.
    :return list: A list with leveraged coins names.
    """
    if not coins_dic:
        coins_dic = get_coins_info_dic(decimal_mode=decimal_mode,
                                       api_key=api_key,
                                       api_secret=api_secret)

    leveraged = []

    coins_up = [i + 'UP' for i in coins_dic.keys()]
    coins_down = [i + 'DOWN' for i in coins_dic.keys()]

    for coin, _ in coins_dic.items():
        if coin in coins_up:
            leveraged.append(coin)
        elif coin in coins_down:
            leveraged.append(coin)
    return leveraged


def get_leveraged_symbols(decimal_mode: bool,
                          api_key: str,
                          api_secret: str,
                          info_dic: dict = None, leveraged_coins: list = None) -> list:
    """
    Search for Binance symbols based on leveraged coins by searching UP or DOWN before an existing coin in symbol,
    leveraged coins examples are:

        .. code-block:: python

            # leveraged coins
            ['1INCHDOWN', '1INCHUP', 'AAVEDOWN', 'AAVEUP', 'ADADOWN', ... ]

            # leveraged symbols

    :param bool decimal_mode: Fixes Decimal return type and operative.
    :param str api_key: A BinPan encrypted in BinPan's way api key. Do not use unencrypted api values, import it from  secret.py file.
    :param str api_secret: A BinPan encrypted in BinPan's way api secret. Do not use unencrypted api values, import it from  secret.py file.
    :param dict info_dic: Avoid fetching the API by passing a dict with symbols data.
    :param list leveraged_coins: Avoid fetching the API for getting coins by passing a list with coins data.
    :return list: A list with leveraged coins names.
    """
    if not info_dic:
        info_dic = get_info_dic()
    if not leveraged_coins:
        leveraged_coins = get_leveraged_coins(decimal_mode=decimal_mode, api_key=api_key, api_secret=api_secret)

    bases = get_bases_dic(info_dic=info_dic)
    quotes = get_quotes_dic(info_dic=info_dic)
    leveraged_symbols = []
    for symbol in info_dic.keys():
        b = bases[symbol]
        q = quotes[symbol]
        if b in leveraged_coins or q in leveraged_coins:
            leveraged_symbols.append(symbol)
    return leveraged_symbols


#######################
# Exchange Statistics #
#######################


def get_quotes_dic(info_dic: dict = None) -> dict:
    """
    Get quote coin each symbol.

    :param dict info_dic: BinPan exchange info dictionary. It's optional to avoid an API call.
    :return dictionary: It gets symbols as keys and quote coin as values.
    """
    if not info_dic:
        info_dic = get_info_dic()
    return {k: v['quoteAsset'] for k, v in info_dic.items()}


def get_bases_dic(info_dic: dict = None) -> dict:
    """
    Get base coin each symbol.

    :param dict info_dic: BinPan exchange info dictionary. It's optional to avoid an API call.
    :return dictionary: It gets symbols as keys and base coin as values.
    :param info_dic:
    :return:
    """
    if not info_dic:
        info_dic = get_info_dic()
    return {k: v['baseAsset'] for k, v in info_dic.items()}


def exchange_status(decimal_mode: bool,
                    api_key: str,
                    api_secret: str,
                    tradeable=True,
                    spot_required=True,
                    margin_required=True,
                    drop_legal=True,
                    filter_leveraged=True,
                    info_dic: dict = None,
                    symbol_filters: dict = None) -> tuple:
    """
    It returns a lot of results: bases_dic, quotes_dic, legal_coins, not_legal_pairs, symbol_filters, filtered_pairs

    :param bool decimal_mode: Fixes Decimal return type and operative.
    :param str api_key: A BinPan encrypted in BinPan's way api key. Do not use unencrypted api values, import it from  secret.py file.
    :param str api_secret: A BinPan encrypted in BinPan's way api secret. Do not use unencrypted api values, import it from  secret.py file.
    :param bool tradeable: Require or not just currently trading symbols.
    :param bool spot_required: Requires just SPOT currently trading symbols.
    :param bool margin_required: Requires just MARGIN currently trading symbols.
    :param bool drop_legal: Drops symbols with legal coins.
    :param bool filter_leveraged: Drops symbols with leveraged coins.
    :param dict info_dic: BinPan exchange info dictionary. It's optional to avoid an API call.
    :param dict symbol_filters: A BinPan symbols filters dict.
    :return tuple: bases_dic, quotes_dic, legal_coins, not_legal_pairs, symbol_filters, filtered_pairs
    """
    # get symbols dict from server with some operational_filters
    if not info_dic:
        info_dic = get_info_dic()

    bases_dic = {k: v['baseAsset'] for k, v in info_dic.items()}
    quotes_dic = {k: v['quoteAsset'] for k, v in info_dic.items()}

    filtered_info = info_dic

    if filter_leveraged:
        filtered_info = filter_leveraged_tokens(info_dic)
    if tradeable:  # elimina los no tradeables
        filtered_info = filter_tradeable(filtered_info)
    if spot_required:
        filtered_info = filter_spot(filtered_info)
    if margin_required:
        filtered_info = filter_margin(filtered_info)

    legal_coins = get_legal_coins(decimal_mode=decimal_mode, api_key=api_key, api_secret=api_secret)
    not_legal_pairs = filter_legal(legal_coins=legal_coins, info_dic=info_dic)
    if drop_legal:
        filtered_info = filter_legal(legal_coins=legal_coins, info_dic=filtered_info)

    if not symbol_filters:
        symbol_filters = get_symbols_filters(info_dic=info_dic)

    filtered_pairs = filtered_info.keys()
    return bases_dic, quotes_dic, legal_coins, not_legal_pairs, symbol_filters, filtered_pairs


def get_24h_statistics(symbol: str = None) -> dict:  # 24h rolling window
    """
    GET /api/v3/ticker/24hr

    24 hour rolling window price change statistics. Careful when accessing this with no symbol.

    Weight(IP):

    Symbols requested weight:


    1-20:	        1

    21-100:	    20

    101 or more:	40

    If symbols parameter is omitted	40

    :param str symbol: Optional symbol.
    :return dict: As example shown.

    Example:

    .. code-block::

       {
          "symbol": "BNBBTC",
          "priceChange": "-94.99999800",
          "priceChangePercent": "-95.960",
          "weightedAvgPrice": "0.29628482",
          "prevClosePrice": "0.10002000",
          "lastPrice": "4.00000200",
          "lastQty": "200.00000000",
          "bidPrice": "4.00000000",
          "bidQty": "100.00000000",
          "askPrice": "4.00000200",
          "askQty": "100.00000000",
          "openPrice": "99.00000000",
          "highPrice": "100.00000000",
          "lowPrice": "0.10000000",
          "volume": "8913.30000000",
          "quoteVolume": "15.30000000",
          "openTime": 1499783499040,
          "closeTime": 1499869899040,
          "firstId": 28385,   // First tradeId
          "lastId": 28460,    // Last tradeId
          "count": 76         // Trade count
        }
    """
    endpoint = '/api/v3/ticker/24hr?'
    if symbol:
        check_weight(endpoint=endpoint, weight=1)
        return api_raw_get(endpoint=endpoint, params={'symbol': symbol}, weight=1)
    else:
        check_weight(endpoint=endpoint, weight=40)
        return api_raw_get(endpoint=endpoint, weight=40)


def not_iterative_coin_conversion(coin: str,
                                  decimal_mode: bool,
                                  prices: dict = None,
                                  try_coin: str = 'BTC',
                                  coin_qty: float = 1) -> float or None:
    """
    Converts any coin quantity value to a reference coin.

    :param bool decimal_mode: Fixes Decimal return type and operative.
    :param str coin: A coin to convert to other coin value.
    :param dict prices: Current prices of all symbols.
    :param str try_coin: Reference coin to convert value to.
    :param float coin_qty: A quantity to use in calculation.
    :return float: Converted value result.
    """
    if not prices:
        prices = get_prices_dic(decimal_mode=decimal_mode)
    if coin + try_coin in prices.keys():
        price = prices[coin + try_coin]
        try_symbol = f"{try_coin}USDT"
        return price * coin_qty * prices[try_symbol]
    elif try_coin + coin in prices.keys():
        price = 1 / prices[try_coin + coin]
        try_symbol = f"USDT{try_coin}"
        try:
            return price * coin_qty * prices[try_symbol]
        except KeyError:
            return None
    else:
        return None


def convert_to_other_coin(coin: str,
                          decimal_mode: bool,
                          convert_to: str = 'USDT',
                          coin_qty: float = 1,
                          prices: dict = None) -> float:
    """
    Convert value of a quantity of coins to value in other coin.

    :param bool decimal_mode: Fixes Decimal return type and operative.
    :param str coin: Your coin.
    :param str convert_to: A Binance existing coin.
    :param float coin_qty: A quantity.
    :param dict prices: A dict with all current prices.
    :return float: Value expressed in the converted coin.
    """
    coin = coin.upper()
    convert_to = convert_to.upper()

    if coin == convert_to:
        return coin_qty

    if not prices:
        prices = get_prices_dic(decimal_mode=decimal_mode)

    symbol_a = coin + convert_to
    symbol_b = convert_to + coin

    if symbol_a in prices.keys():
        # print(f"{coin} Selected {symbol_a} for ", coin)
        return coin_qty * prices[symbol_a]

    elif symbol_b in prices.keys():
        # print(f"{coin} Selected {symbol_b} for ", coin)
        return coin_qty * (1 / prices[symbol_b])
    else:
        # try using btc
        # try:
        ret1 = not_iterative_coin_conversion(coin=coin, prices=prices, try_coin='BTC', coin_qty=coin_qty, decimal_mode=decimal_mode)
        if ret1:
            return ret1
        ret2 = not_iterative_coin_conversion(coin=coin, prices=prices, try_coin='BUSD', coin_qty=coin_qty, decimal_mode=decimal_mode)
        if ret2:
            return ret2
        ret3 = not_iterative_coin_conversion(coin=coin, prices=prices, try_coin='BNB', coin_qty=coin_qty, decimal_mode=decimal_mode)
        if ret3:
            return ret3
        ret4 = not_iterative_coin_conversion(coin=coin, prices=prices, try_coin='ETH', coin_qty=coin_qty, decimal_mode=decimal_mode)
        if ret4:
            return ret4
        ret5 = not_iterative_coin_conversion(coin=coin, prices=prices, try_coin='TUSD', coin_qty=coin_qty, decimal_mode=decimal_mode)
        if ret5:
            return ret5
        ret6 = not_iterative_coin_conversion(coin=coin, prices=prices, try_coin='USDC', coin_qty=coin_qty, decimal_mode=decimal_mode)
        if ret6:
            return ret6
        else:
            return np.nan


def convert_symbol_base_to_other_coin(decimal_mode: bool,
                                      symbol_to_convert_base: str = 'ETHBTC',
                                      base_qty: float = 1,
                                      convert_to: str = 'USDT',
                                      prices: dict = None,
                                      info_dic: dict = None) -> float or dd:
    """
        Convert value of a quantity of coins to value in other coin.

    :param bool decimal_mode: Fixes Decimal return type and operative.
    :param str symbol_to_convert_base: A symbol to get it's base to convert to other coin value.
    :param float base_qty: A quantity.
    :param str convert_to: A Binance existing coin.
    :param dict prices: A dict with all current prices.
    :param dict info_dic: BinPan exchange info dictionary. It's optional to avoid an API call.
    :return float: Value expressed in the converted coin.
    """
    if decimal_mode:
        my_type = dd
    else:
        my_type = float

    if not info_dic:
        info_dic = get_info_dic()

    # bases_dic = {k: v['baseAsset'] for k, v in info_dic.items()}
    bases_dic = get_bases_dic(info_dic=info_dic)
    try:
        base = bases_dic[symbol_to_convert_base]
    except KeyError:
        return 0

    if not prices:
        prices = get_prices_dic(decimal_mode=decimal_mode)
    ret = convert_to_other_coin(coin=base, convert_to=convert_to, coin_qty=base_qty, prices=prices, decimal_mode=decimal_mode)
    if type(ret) == int or type(ret) == float or type(ret) == dd:
        return my_type(ret)
    else:
        raise Exception('BinPan Error: convert_symbol_base_to_other_coin breakpoint')


def convert_utc_milliseconds(ms: int) -> str:
    """
    Converts milliseconds timestamp to formatted string.

    :param int ms: Milliseconds linux timestamp from epoch.
    :return str: Formatted date.
    """
    seconds = int(ms) / 1000
    return str(datetime.utcfromtimestamp(seconds).strftime('%Y-%m-%d %H:%M:%S.%f'))


def statistics_24h(decimal_mode: bool,
                   api_key: str,
                   api_secret: str,
                   tradeable=True,
                   spot_required=True,
                   margin_required=False,
                   drop_legal=True,
                   filter_leveraged=True,
                   info_dic=None,
                   stablecoin_value='BUSD',
                   sort_by: str = 'priceChangePercent') -> pd.DataFrame:
    """
    Generates a dataframe with the filters to apply with the statistics of the last 24 hours. Optionally, you can generate the column to
    convert the volume to USDT.

    :param bool decimal_mode: Fixes Decimal return type and operative.
    :param str api_key: A BinPan encrypted in BinPan's way api key. Do not use unencrypted api values, import it from  secret.py file.
    :param str api_secret: A BinPan encrypted in BinPan's way api secret. Do not use unencrypted api values, import it from  secret.py file.
    :param bool tradeable: Require or not just currently trading symbols.
    :param bool spot_required: Requires just SPOT currently trading symbols.
    :param bool margin_required: Requires just MARGIN currently trading symbols.
    :param bool drop_legal: Drops symbols with legal coins.
    :param bool filter_leveraged: Drops symbols with leveraged coins.
    :param dict info_dic: BinPan exchange info dictionary. It's optional to avoid an API call.
    :param stablecoin_value: StableCoin reference for value.
    :param str sort_by: A column to sort by. Default is 'priceChangePercent'.
    :return pd.DataFrame: Picture result.

    .. image:: images/exchange_statistics_24h.png
       :width: 1000
       :alt: exchange_statistics_24h.png

    """

    if not info_dic:
        info_dic = get_info_dic()

    bases_dic, quotes_dic, _, not_legal_pairs, _, filtered_pairs = exchange_status(tradeable=tradeable,
                                                                                   spot_required=spot_required,
                                                                                   margin_required=margin_required,
                                                                                   drop_legal=drop_legal,
                                                                                   filter_leveraged=filter_leveraged,
                                                                                   info_dic=info_dic,
                                                                                   decimal_mode=decimal_mode,
                                                                                   api_key=api_key,
                                                                                   api_secret=api_secret)
    stats = get_24h_statistics()

    df = pd.DataFrame(stats)

    # filter just coins not legal and with spot and tradeable
    df = df[df['symbol'].isin(filtered_pairs)]

    df = df.apply(pd.to_numeric, errors='ignore')

    df['openTime'] = df['openTime'].apply(lambda x: convert_utc_milliseconds(x))
    df['closeTime'] = df['closeTime'].apply(lambda x: convert_utc_milliseconds(x))

    df = df.set_index('symbol', drop=False)

    # df['filter_passed'] = df['symbol'].apply(lambda x: True if x in filtered_pairs else False)
    df['is_legal'] = df['symbol'].apply(lambda x: True if x not in not_legal_pairs else False)

    df['base'] = df['symbol'].apply(lambda x: bases_dic[x])
    df['quote'] = df['symbol'].apply(lambda x: quotes_dic[x])

    if stablecoin_value:
        prices = get_prices_dic(decimal_mode=decimal_mode)
        # info_dic = {k['symbol']: k for k in market.get_exchange_info()['symbols']}
        # bases_dic = market.get_bases_dic(info_dic=info_dic)
        # quotes_dic = market.get_quotes_dic(info_dic=info_dic)

        stable_coin_value_name = f"{stablecoin_value}_value"

        for pair in prices.keys():
            usdt_val = convert_symbol_base_to_other_coin(symbol_to_convert_base=pair,
                                                         base_qty=1,
                                                         convert_to=stablecoin_value,
                                                         prices=prices,
                                                         info_dic=info_dic,
                                                         decimal_mode=decimal_mode)

            df.loc[df['symbol'] == pair, stable_coin_value_name] = float(usdt_val)  # float for pandas

        stable_coin_volumen_name = f"{stablecoin_value}_volume"
        df[stable_coin_volumen_name] = df[stable_coin_value_name] * df['volume']

        return df.sort_values(stable_coin_volumen_name, ascending=False)

    return df.sort_values(sort_by, ascending=False)


def get_top_gainers(decimal_mode: bool,
                    api_key: str,
                    api_secret: str,
                    info_dic: dict = None,
                    tradeable=True,
                    spot_required=True,
                    margin_required=False,
                    drop_legal=True,
                    filter_leveraged=True,
                    top_gainers_qty: int = None,
                    my_quote: str = 'BUSD',
                    drop_stable_pairs: bool = True,
                    sort_by_column: str = 'priceChangePercent',
                    full_return: bool = False
                    ) -> pd.DataFrame:
    """
    Generates a dataframe for symbols against a quote with the filters to apply with the statistics of the last 24 hours.
    Optionally, you can generate the column to convert the volume to USDT.

    :param bool decimal_mode: Fixes Decimal return type and operative.
    :param str api_key: A BinPan encrypted in BinPan's way api key. Do not use unencrypted api values, import it from  secret.py file.
    :param str api_secret: A BinPan encrypted in BinPan's way api secret. Do not use unencrypted api values, import it from  secret.py file.
    :param bool tradeable: Require or not just currently trading symbols.
    :param bool spot_required: Requires just SPOT currently trading symbols.
    :param bool margin_required: Requires just MARGIN currently trading symbols.
    :param bool drop_legal: Drops symbols with legal coins.
    :param bool filter_leveraged: Drops symbols with leveraged coins.
    :param dict info_dic: BinPan exchange info dictionary. It's optional to avoid an API call.
    :param top_gainers_qty: Limit result top lines.
    :param str my_quote: A quote coin to reference values.
    :param bool drop_stable_pairs: It drop stablecoin to stablecoin symbols.
    :param str sort_by_column: A column to sort by. Default is 'priceChangePercent'.
    :param full_return: Activate return full columns data from exchange, else, returns just a few basic columns.
    :return pd.DataFrame: A dataframe as in the example .

    Example:

    .. code-block::

        from binpan import handlers
        from secret import api_key, api_secret

        handlers.exchange.get_top_gainers(decimal_mode=False, api_key=api_key, api_secret=api_secret)

              priceChangePercent	volume	BUSD_value	BUSD_volume
        symbol
        PHABTC	198.477	    9.291796e+07	0.33650	3.126689e+07
        PHAUSDT	195.961	    6.729636e+08	0.33650	2.264522e+08
        PHABUSD	195.530	    1.082033e+09	0.33650	3.641039e+08
        LTOBTC	 61.732	    5.854178e+07	0.11070	6.480575e+06
        LTOUSDT	 60.637	    2.813419e+08	0.11070	3.114455e+07
        ...	...	...	...	    ...
        MDXBTC	 -21.195	4.096356e+07	0.15640	6.406700e+06
        MDXBUSD	 -22.112	3.444397e+08	0.15640	5.387037e+07
        MDXUSDT	 -22.123	2.896132e+08	0.15640	4.529550e+07
        AGIXBTC	 -23.007	1.086561e+08	0.06489	7.050693e+06
        AGIXBUSD -23.407	2.188445e+08	0.06489	1.420082e+07
        1194 rows × 4 columns

    """
    if not info_dic:
        info_dic = get_info_dic()

    top_gainers = statistics_24h(decimal_mode=decimal_mode,
                                 api_key=api_key,
                                 api_secret=api_secret,
                                 tradeable=tradeable,
                                 spot_required=spot_required,
                                 margin_required=margin_required,
                                 drop_legal=drop_legal,
                                 info_dic=info_dic,
                                 filter_leveraged=filter_leveraged,
                                 stablecoin_value=my_quote)
    # # filter quote
    # top_gainers = top_gainers.loc[top_gainers['quote'] == my_quote]

    if drop_stable_pairs:
        stable_pairs = [f"{s}{my_quote}" for s in stablecoins]
        stable_pairs += [f"{my_quote}{s}" for s in stablecoins]

        top_gainers = top_gainers.loc[~top_gainers['symbol'].isin(stable_pairs)]

    if not full_return:
        top_gainers = top_gainers[['priceChangePercent', 'volume', f'{my_quote}_value', f'{my_quote}_volume']]

    if top_gainers_qty:
        return top_gainers.sort_values(sort_by_column, ascending=False).head(top_gainers_qty)
    else:
        return top_gainers.sort_values(sort_by_column, ascending=False)


#################
# numeric tools #
#################


def get_decimal_positions(num: float or dd) -> int:
    """
    Count decimal positions for a value.

    :param float or dd num: Input.
    :return int: Count decimal positions.
    """
    num = dd(num).normalize()
    return abs(num.as_tuple().exponent)



# coding=utf-8
"""

API requests module.

"""
from typing import List, Tuple

from .logs import Logs

# from .exceptions import BinanceAPIException, BinanceRequestException
from .exceptions import *


from .starters import AesCipher, get_exchange_limits

from urllib.parse import urljoin, urlencode
import requests
import hmac
import hashlib
import copy
from time import sleep


float_api_items = ['price', 'origQty', 'executedQty', 'cummulativeQuoteQty', 'stopLimitPrice', 'stopPrice', 'commission', 'qty',
                   'origQuoteOrderQty', 'makerCommission', 'takerCommission']
int_api_items = ['orderId', 'orderListId', 'transactTime', 'tradeId', 'transactionTime', 'updateTime', 'time']

base_url = 'https://api.binance.com'

tick_seconds = {'1m': 60, '3m': 60 * 3, '5m': 5 * 60, '15m': 15 * 60, '30m': 30 * 60, '1h': 60 * 60, '2h': 60 * 60 * 2,
                '4h': 60 * 60 * 4, '6h': 60 * 60 * 6, '8h': 60 * 60 * 8, '12h': 60 * 60 * 12, '1d': 60 * 60 * 24,
                '3d': 60 * 60 * 24 * 3, '1w': 60 * 60 * 24 * 7, '1M': 60 * 60 * 24 * 30}

tick_interval_values = ['1m', '3m', '5m', '15m', '30m', '1h', '2h', '4h', '6h', '8h', '12h', '1d', '3d', '1w', '1M']

weight_logger = Logs(filename='./logs/weight.log', name='weight', info_level='INFO')
quest_logger = Logs(filename='./logs/quest.log', name='quest', info_level='INFO')

cipher_object = AesCipher()

# rate limits
api_rate_limits = get_exchange_limits()

current_weight = {}
endpoint_headers = {}  # read temp file

aplicable_limits = {'X-SAPI-USED-IP-WEIGHT-1M': api_rate_limits['REQUEST_1M'],
                    'X-SAPI-USED-UID-WEIGHT-1M': api_rate_limits['REQUEST_1M'],
                    'x-mbx-used-weight': api_rate_limits['REQUEST_5M'],
                    'x-mbx-used-weight-1m': api_rate_limits['REQUEST_1M'],
                    'x-mbx-order-count-10s': api_rate_limits['ORDERS_10S'],
                    'x-mbx-order-count-1d': api_rate_limits['ORDERS_1D']}

api_limits_weight_decrease_per_seconds = {'X-SAPI-USED-IP-WEIGHT-1M': api_rate_limits['REQUEST_1M'] // 60,
                                          'X-SAPI-USED-UID-WEIGHT-1M': api_rate_limits['REQUEST_1M'] // 60,
                                          'x-mbx-used-weight-1m': api_rate_limits['REQUEST_1M'] // 60,
                                          'x-mbx-used-weight': api_rate_limits['REQUEST_5M'] // (60 * 5),
                                          'x-mbx-order-count-10s': api_rate_limits['ORDERS_10S'] // 10,
                                          'x-mbx-order-count-1d': api_rate_limits['ORDERS_1D'] // (
                                                  24 * 60 * 60)}  # is the five minutes api limit?


# TODO: identify new headers for order endpoints

#################
# Aux functions #
#################


def add_header_for_endpoint(endpoint: str, header: str):
    """
    Adds API headers response info for weight control.

    :param str endpoint: Endpoint path.
    :param str header: Header in response.
    :return: None
    """
    global endpoint_headers

    if endpoint in endpoint_headers.keys():
        my_headers = endpoint_headers[endpoint]
        if header not in my_headers:
            my_headers.append(header)
            endpoint_headers.update({endpoint: my_headers})
    else:
        endpoint_headers.update({endpoint: [header]})
        weight_logger.debug(f"New endpoint headers control: {endpoint} -> {endpoint_headers[endpoint]}")


def update_weights(headers, father_url: str):
    """
    Update weights on headers weight control global dictionary, used in control of pauses in API requests to avoid bans.

    :param headers: Response headers. Usually a dictionary.
    :param str father_url: API url requested.
    :return: None
    """
    global current_weight

    weight_logger.debug(f"Header: {headers}")

    # get weight headers from headers
    weight_headers = {k: int(v) for k, v in headers.items() if 'WEIGHT' in k.upper() or 'COUNT' in k.upper()}
    weight_logger.debug(f"Father url: {father_url} Weights updated from API: {weight_headers}")

    for head, value in weight_headers.items():
        add_header_for_endpoint(father_url, head)

    # decrease all other weights
    for k, v in current_weight.items():
        new_v = max(0, v - 1)  # is possible to decrease too much not used endpoints when using a variety of them
        current_weight.update({k: new_v})

    # annotate new weights updated
    for k, v in weight_headers.items():
        current_weight.update({k: v})

    weight_logger.debug(f"Current weights updated: {current_weight}")


def check_weight(weight: int,
                 endpoint: str):
    """
    Verify what weight will result before requesting API and compare to API limits.

    :param int weight: Weight expected from the request planned.
    :param str endpoint: Endpoint requested to find API weight.
    :return: None
    """
    global current_weight, endpoint_headers, aplicable_limits, api_limits_weight_decrease_per_seconds

    weight_logger.debug(f"Checking weight for {endpoint}")

    future_weights = copy.deepcopy(current_weight)

    if not future_weights:  # cold start
        return
    weight_logger.debug(f"Future weight headers: {future_weights}")

    # añade los headers nuevos al archivo de endpoint headers
    for future_key in future_weights.keys():
        add_header_for_endpoint(endpoint, future_key)

    # busca los headers a incrementar usando el archivo de endpoint headers
    for endpoint_archived, end_headers_list in endpoint_headers.items():
        if endpoint_archived == endpoint:
            for head in end_headers_list:
                future_weights[head] += weight

    # check if needed header for this endpoint is overloaded
    aplicable_headers = endpoint_headers[endpoint]
    for aplicable_head in aplicable_headers:
        curr_limit = aplicable_limits[aplicable_head]  # must raise if new header in limits control
        future_weight = future_weights[aplicable_head]
        if future_weight >= curr_limit:
            excess = ((future_weight - curr_limit) // api_limits_weight_decrease_per_seconds[aplicable_head]) + 2

            weight_logger.warning(
                f"{aplicable_head}={future_weight} > {aplicable_limits[aplicable_head]} Current value: {current_weight[aplicable_head]}")
            weight_logger.warning(f"Waiting {excess} seconds for {aplicable_head} to decrease rate.")
            sleep(excess)


def get_server_time() -> int:
    """
    Get time from server.

    :return int: A linux timestamp in milliseconds.
    """
    endpoint = '/api/v3/time'
    check_weight(1, endpoint=endpoint)
    return int(get_response(url=endpoint)['serverTime'])


##################
# ## Requests ## #
##################


def convert_response_type(response_data: dict or list,
                          decimal_mode: bool) -> dict or list:
    """
    Infers types into api response and changes those.

    :param dict or list response_data: API response json loaded.
    :param bool decimal_mode: Sets numeric data type to decimal.
    :return dict or list: Typed API data in response..
    """
    if decimal_mode:
        return response_data

    if type(response_data) == dict:
        response = response_data.copy()
        for k, v in response.items():
            if k in float_api_items:
                response[k] = float(v)
            elif k in int_api_items:
                response[k] = int(v)
            elif type(v) == dict:
                response[k] = convert_response_type(v, decimal_mode=decimal_mode)
            elif type(v) == list:
                response[k] = [convert_response_type(ii, decimal_mode=decimal_mode) for ii in v]
    elif type(response_data) == list:
        response = [convert_response_type(iii, decimal_mode=decimal_mode) for iii in response_data]
    else:
        response = response_data
    return response


def get_response(url: str, params: dict or List[tuple] = None, headers: dict = None) -> dict or list:
    """
    Requests GET to API. Before requesting calculates resulting weight and waits enough time to not surplus limit for that endpoint.

    :param str url: API endpoint url.
    :param dit or List[tuple] params: Request params.
    :param dict headers: Request headers.
    :return dict or list: API response in dict or list format.
    """
    if not url.startswith(base_url):
        url = urljoin(base_url, url)

    response = requests.get(url, params=params, headers=headers)
    update_weights(headers=response.headers,
                   father_url=url)
    quest_logger.debug(f"get_response parameters: {locals()}")

    return handle_api_response(response)


def post_response(url: str, params: dict or List[tuple] = None, headers: dict = None) -> dict or list:
    """
    Requests POST to API. Before requesting calculates resulting weight and waits enough time to not surplus limit for that endpoint.

    :param str url: API endpoint url.
    :param dit or List[tuple] params: Request params.
    :param dict headers: Request headers.
    :return: API response in dict or list format.
    """
    quest_logger.debug(f"post_response: params: {params} headers: {headers}")
    if not url.startswith(base_url):
        url = urljoin(base_url, url)
    response = requests.post(url, params=params, headers=headers)
    update_weights(response.headers, father_url=url)
    quest_logger.debug(f"post_response: response: {response}")
    return handle_api_response(response)


def delete_response(url: str, params: dict or List[tuple] = None, headers: dict = None) -> dict or list:
    """
    Requests DELETE to API. Before requesting calculates resulting weight and waits enough time to not surplus limit for that endpoint.

    :param str url: API endpoint url.
    :param dit or List[tuple] params: Request params.
    :param dict headers: Request headers.
    :return: API response in dict or list format.
    """
    if not url.startswith(base_url):
        url = urljoin(base_url, url)
    response = requests.delete(url, params=params, headers=headers)
    update_weights(response.headers, father_url=url)
    quest_logger.debug(f"delete_response: response: {response}")
    return handle_api_response(response)


def handle_api_response(response) -> dict or list:
    """
    Raises if there is any problem with the server response or returns the raw json.

    :return: API response in dict or list format.
    """
    if not (200 <= response.status_code < 300):
        quest_logger.error(response)
        raise BinanceAPIException(response, response.status_code, response.text)
    try:
        return response.json()
    except ValueError:
        quest_logger.error(response)
        raise BinanceRequestException(f'Invalid Response: {response.text}')


def hashed_signature(url_params: str,
                     api_secret: str) -> object:
    """
    Hashes params of a request with encoded API secret. Decodes it and return correct hashed signature.

    :param str url_params: String of params to encode.
    :param str api_secret: Encoded API secret.
    :return: Hashed signature.
    """
    return hmac.new(cipher_object.decrypt(api_secret).encode('utf-8'), url_params.encode('utf-8'),
                    hashlib.sha256).hexdigest()


def sign_request(params: dict or List[tuple],
                 recvWindow: int or None,
                 api_key: str,
                 api_secret: str
                 ) -> Tuple[list, dict]:
    """
    Add signature to the request. Returns a list of params in tuples and a headers dict.

    :param dict or List[tuple] params: Params for the request.
    :param int or None recvWindow: Milliseconds of life for the request to be responded. /api/v3/historicalTrades do not admit it.
    :param str api_key: Encoded API key.
    :param str api_secret: Encoded API secret.
    :return Tuple[list, dict]: List of params in tuples and a headers dict.
    """
    quest_logger.debug(f"parse_request: {params}")

    if params is None:
        params = {}

    # clean none params
    params.update({'recvWindow': recvWindow})
    params = {k: v for k, v in params.items() if v is not None}

    params_tuples = []
    for k, v in params.items():
        if type(v) != list:
            params_tuples.append((k, v,))
        else:
            for i in v:
                params_tuples.append((k, i,))

    key = cipher_object.decrypt(api_key)
    headers = {"X-MBX-APIKEY": key}

    server_time_int = get_server_time()
    params_tuples.append(("timestamp", server_time_int,))  # añado el timestamp como parámetro
    signature = hashed_signature(urlencode(params_tuples), api_secret=api_secret)
    params_tuples.append(("signature", signature,))  # y la añado como parámetro

    return params_tuples, headers


def get_signed_request(url: str,
                       decimal_mode: bool,
                       api_key: str,
                       api_secret: str,
                       params: dict or List[tuple] = None,
                       recvWindow: int = 10000
                       ) -> dict or list:
    """
    It does a signed get to a url along with a dictionary of parameters. To avoid parameter order errors, they are passed in tuple format,
    it is also useful sending parameters that have the same name with several values:

        https://dev.binance.vision/t/faq-signature-for-this-request-is-not-valid/176/4

    :param str url: API endpoint.
    :param bool decimal_mode: Sets numeric data format to decimal.
    :param str api_key: Encoded API key.
    :param str api_secret: Encoded API secret.
    :param dict or List[tuple] params: Params for the request.
    :param int recvWindow: Milliseconds of life for the request to be responded.
    :return dict or list: API response in dict or list format.
    """
    params_tuples, headers = sign_request(params=params, recvWindow=recvWindow, api_key=api_key, api_secret=api_secret)
    ret = get_response(url, params=params_tuples, headers=headers)
    quest_logger.debug("get_signed_request: params_tuples: " + str(params_tuples))
    quest_logger.debug("get_signed_request: headers: " + str(headers.keys()))
    return convert_response_type(ret, decimal_mode=decimal_mode)


def get_semi_signed_request(url: str,
                            decimal_mode: bool,
                            api_key: str,
                            params: dict = None) -> dict or list:
    """
    Requests GET with api key header and params. Some methods requested this way.

    :param str url: API endpoint.
    :param bool decimal_mode: Sets numeric data format to decimal.
    :param str api_key: Encoded API key.
    :param dict or List[tuple] params: Params for the request.
    :return dict or list: API response in dict or list format.
    """
    headers = {"X-MBX-APIKEY": cipher_object.decrypt(api_key)}
    ret = get_response(url=url, params=params, headers=headers)
    quest_logger.debug("get_semi_signed_request: headers: " + str(headers.keys()))
    return convert_response_type(ret, decimal_mode=decimal_mode)


def post_signed_request(url: str,
                        decimal_mode: bool,
                        api_key: str,
                        api_secret: str,
                        params: dict = None,
                        recvWindow: int = 10000) -> dict or list:
    """
    Makes a signed POST to a url along with a dictionary of parameters.

    :param str url: API endpoint.
    :param bool decimal_mode: Sets numeric data format to decimal.
    :param str api_key: Encoded API key.
    :param str api_secret: Encoded API secret.
    :param dict or List[tuple] params: Params for the request.
    :param int recvWindow: Milliseconds of life for the request to be responded.
    :return dict or list: API response in dict or list format.
    """
    params_tuples, headers = sign_request(params=params, recvWindow=recvWindow, api_key=api_key, api_secret=api_secret)
    ret = post_response(url, params=params_tuples, headers=headers)
    quest_logger.debug("post_signed_request: params_tuples: " + str(params_tuples))
    quest_logger.debug("get_semi_signed_request: headers: " + str(headers.keys()))
    return convert_response_type(ret, decimal_mode=decimal_mode)


def delete_signed_request(url: str,
                          decimal_mode: bool,
                          api_key: str,
                          api_secret: str,
                          params: dict = None,
                          recvWindow: int = 10000) -> dict or list:
    """
    Makes a signed DELETE to a url along with a dictionary of parameters.

    :param str url: API endpoint.
    :param bool decimal_mode: Sets numeric data format to decimal.
    :param str api_key: Encoded API key.
    :param str api_secret: Encoded API secret.
    :param dict or List[tuple] params: Params for the request.
    :param int recvWindow: Milliseconds of life for the request to be responded.
    :return dict or list: API response in dict or list format.
    """
    params_tuples, headers = sign_request(params=params, recvWindow=recvWindow, api_key=api_key, api_secret=api_secret)
    ret = delete_response(url, params=params_tuples, headers=headers)
    quest_logger.info("delete_signed_request: params_tuples: " + str(params_tuples))
    quest_logger.debug("get_semi_signed_request: headers: " + str(headers.keys()))
    return convert_response_type(ret, decimal_mode=decimal_mode)


####################
# Generic Requests #
####################

def api_raw_get(endpoint: str,
                weight: int,
                base_url: str = '',
                params: dict or List[tuple] = None,
                headers: dict = None) -> dict or list:
    """
    Shortcut to request GET to API.

    :param str endpoint: API endpoint to request.
    :param int weight: Expected weight for the request.
    :param str base_url: Base URL.
    :param dict or List[tuple] params: Params for the request.
    :param dict headers: Headers to use in the request.
    :return dict or list: API response in dict or list format.
    """
    check_weight(weight, endpoint=base_url + endpoint)

    return get_response(url=base_url + endpoint,
                        params=params,
                        headers=headers)


def api_raw_signed_get(endpoint: str,
                       decimal_mode: bool,
                       api_key: str,
                       api_secret: str,
                       base_url: str = '',
                       params: dict or List[tuple] = None,
                       weight: int = 1) -> dict or list:
    """
    Shortcut to request signed GET to API.

    :param bool decimal_mode: If True, declares format of numbers to decimal type.
    :param str api_key: Encoded API key.
    :param str api_secret: Encoded API secret.
    :param str endpoint: API endpoint to request.
    :param int weight: Expected weight for the request.
    :param str base_url: Base URL.
    :param dict or List[tuple] params: Params for the request.
    :return dict or list: API response in dict or list format.
    """
    check_weight(weight, endpoint=base_url + endpoint)
    return get_signed_request(url=base_url + endpoint,
                              params=params, decimal_mode=decimal_mode, api_key=api_key, api_secret=api_secret)


def api_raw_signed_post(endpoint: str,
                        decimal_mode: bool,
                        api_key: str,
                        api_secret: str,
                        base_url: str = '',
                        params: dict or List[tuple] = None,
                        weight: int = 1) -> dict or list:
    """
    Shortcut to request signed POST to API.

    :param bool decimal_mode: If True, declares format of numbers to decimal type.
    :param str api_key: Encoded API key.
    :param str api_secret: Encoded API secret.
    :param str endpoint: API endpoint to request.
    :param int weight: Expected weight for the request.
    :param str base_url: Base URL.
    :param dict or List[tuple] params: Params for the request.
    :return dict or list: API response in dict or list format.
    """
    check_weight(weight, endpoint=base_url + endpoint)
    return post_signed_request(url=base_url + endpoint,
                               params=params, decimal_mode=decimal_mode, api_key=api_key, api_secret=api_secret)
